#include <asm/uaccess.h>    /* copy_from_user, copy_to_user */
#include <linux/errno.h>
#include <linux/fs.h>       /* inode, file, file_operations */
#include <linux/kernel.h>   /* printk */
#include <linux/module.h>
#include <linux/spinlock.h>


MODULE_DESCRIPTION( "Sample character device driver." );
MODULE_LICENSE( "GPL" );


static int   devmajor = 242;
static char* devname  = "chardev";
static char* msg      = "module [chardev.o]";


#define MAXBUF 32
static unsigned char devbuf[ MAXBUF ];
static int buf_pos;

static int access_num;
static spinlock_t chardev_spin_lock;


/*
 * open()
 * 1桼ץΤߤǥХե open 롣
 */
static int 
chardev_open( struct inode* inode, struct file* filp )
{
	printk( KERN_INFO "%s : open()  called\n", msg );

	spin_lock( &chardev_spin_lock );

	if ( access_num ) {
		spin_unlock( &chardev_spin_lock );
		return -EBUSY;
	}

	access_num ++;
	spin_unlock( &chardev_spin_lock );

	return 0;
}


/*
 * release()
 * ѥ桼ץ򥯥ꥢ롣
 */
static int 
chardev_release( struct inode* inode, struct file* filp )
{
	printk( KERN_INFO "%s : close() called\n", msg );

	spin_lock( &chardev_spin_lock );
	access_num --;
	spin_unlock( &chardev_spin_lock );

	return 0;
}


/*
 * read()
 * 桼ץФХåեƤž롣
 * žƤХåեä롣
 */
static ssize_t 
chardev_read( struct file* filp, char* buf, size_t count, loff_t* pos )
{
	int copy_len;
	int i;

	printk( KERN_INFO "%s : read()  called\n", msg );

	if ( count > buf_pos )
		copy_len = buf_pos;
	else
		copy_len = count;

	if ( copy_to_user( buf, devbuf, copy_len ) ) {
		printk( KERN_INFO "%s : copy_to_user failed\n", msg );
		return -EFAULT;
	}

	*pos += copy_len;

	for ( i = copy_len; i < buf_pos; i ++ )
		devbuf[ i - copy_len ] = devbuf[i];

	buf_pos -= copy_len;
	printk( KERN_INFO "%s : buf_pos = %d\n", msg, buf_pos );

	return copy_len;
}


/*
 * write()
 * 桼ץž줿ƤХåե˽񤭹ࡣ
 */
static ssize_t 
chardev_write( struct file* filp, const char* buf, size_t count, loff_t* pos )
{
	int copy_len;

	printk( KERN_INFO "%s : write() called\n", msg );

	if ( buf_pos == MAXBUF ) {
		printk( KERN_INFO "%s : no space left\n", msg );
		return -ENOSPC;
	}

	if ( count > ( MAXBUF - buf_pos ) )
		copy_len = MAXBUF - buf_pos;
	else
		copy_len = count;

	if ( copy_from_user( devbuf + buf_pos, buf, copy_len ) ) {
		printk( KERN_INFO "%s : copy_from_user failed\n", msg );
		return -EFAULT;
	}

	*pos    += copy_len;
	buf_pos += copy_len;

	printk( KERN_INFO "%s : buf_pos = %d\n", msg, buf_pos );

	return copy_len;
}


static struct file_operations chardev_fops = 
{
	owner   : THIS_MODULE,
	read    : chardev_read,
	write   : chardev_write,
	open    : chardev_open,
	release : chardev_release,
};


/*
 * ⥸塼ν
 * insmod ˸ƤФ
 */
int 
init_module( void )
{
	if ( register_chrdev( devmajor, devname, &chardev_fops ) ) {
		printk( KERN_INFO "%s : register_chrdev failed\n" );
		return -EBUSY;
	}

	spin_lock_init( &chardev_spin_lock );
	printk( KERN_INFO "%s : loaded  into kernel\n", msg );

	return 0;
}


/*
 * ⥸塼β
 * ⥸塼λȿ 0 ǤСrmmod ˸ƤФ
 */
void 
cleanup_module( void )
{
	if ( unregister_chrdev( devmajor, devname ) ) {
		printk( KERN_INFO "%s : unregister_chrdev failed\n" );
		/* к̵ ... */
	}

	printk( KERN_INFO "%s : removed from kernel\n", msg );
}


/* End of chardev.c */
